""" Test helpers for vxfreeswitch. """

from uuid import uuid4

from zope.interface import implements

import time

from twisted.internet import reactor
from twisted.internet.defer import (
    inlineCallbacks, returnValue, DeferredQueue, Deferred)
from twisted.internet.endpoints import TCP4ServerEndpoint
from twisted.internet.protocol import Protocol, Factory, ClientFactory
from twisted.protocols.basic import LineReceiver
from twisted.test.proto_helpers import StringTransport

from vumi.tests.helpers import IHelper, proxyable


class EslCommand(object):
    """
    An object representing an ESL command.
    """
    def __init__(self, cmd_type, params=None):
        self.cmd_type = cmd_type
        self.params = params if params is not None else {}

    def __repr__(self):
        return "<%s cmd_type=%r params=%r>" % (
            self.__class__.__name__, self.cmd_type, self.params)

    def __eq__(self, other):
        if not isinstance(other, EslCommand):
            return NotImplemented
        return (self.cmd_type == other.cmd_type and
                self.params == other.params)

    def __getitem__(self, name):
        return self.params.get(name)

    def __setitem__(self, name, value):
        self.params[name] = value

    @classmethod
    def from_dict(cls, d):
        """
        Convert a dict to an :class:`EslCommand`.
        """
        cmd_type = d.get("type")
        params = {
            "call-command": d.get("call-command", "execute"),
            "event-lock": d.get("event-lock", "true"),
        }
        if "name" in d:
            params["execute-app-name"] = d.get("name")
        if "arg" in d:
            params["execute-app-arg"] = d.get("arg")
        return cls(cmd_type, params)


class EslParser(object):
    """
    Simple in-efficient parser for the FreeSwitch eventsocket protocol.
    """

    def __init__(self):
        self.data = ""

    def parse(self, new_data):
        data = self.data + new_data
        cmds = []
        while "\n\n" in data:
            cmd_data, data = data.split("\n\n", 1)
            command = EslCommand("unknown")
            first_line = True
            for line in cmd_data.splitlines():
                line = line.strip()
                if not line:
                    continue
                if first_line:
                    command.cmd_type = line.strip()
                    first_line = False
                    continue
                if ":" in line:
                    key, value = line.split(":", 1)
                    command[key] = value.strip()
            cmds.append(command)
        self.data = data
        return cmds


class EslTransport(StringTransport):

    def __init__(self):
        StringTransport.__init__(self)
        self.cmds = DeferredQueue()
        self.esl_parser = EslParser()

    def write(self, data):
        StringTransport.write(self, data)
        for cmd in self.esl_parser.parse(data):
            self.cmds.put(cmd)


class RecordingServer(Protocol):
    def __init__(self):
        self.command_parser = EslParser()

    def connectionMade(self):
        self.factory.clients.append(self)

    def _send_event(self, content):
        self.transport.write(
            'Content-Length: %s\n' % len(content) +
            'Content-Type: text/event-plain\n\n' +
            content)

    def dataReceived(self, line):
        commands = self.command_parser.parse(line)
        for cmd in commands:
            response = self.factory.get_response(cmd)
            self.transport.write(response.to_bytes())

    def hangup(self):
        content = (
            'Event-Name: CHANNEL_HANGUP\n')
        self._send_event(content)


class FixtureResponse(object):
    """ A response to an ESL command. """

    AUTH_REQUEST = 'auth/request'
    API_RESPONSE = 'api/response'
    REPLY = 'command/reply'
    EVENT = 'text/event-plain'

    def __init__(self, content_type, content=None, headers=()):
        self.content_type = content_type
        self.content = content
        self.headers = headers

    def to_bytes(self):
        lines = []
        lines.append("Content-Type: %s\n" % self.content_type)
        if self.content is not None:
            lines.append("Content-Length: %d\n" % len(self.content))
        lines.extend("%s: %s\n" % (k, v) for k, v in self.headers)
        lines.append("\n")
        if self.content is not None:
            lines.append(self.content)
        return "".join(lines)


class FixtureReply(FixtureResponse):
    """ A reply to an ESL command. """

    def __init__(self, *args):
        headers = [("Reply-Text", " ".join(args))]
        super(FixtureReply, self).__init__(self.REPLY, headers=headers)


class FixtureAuthResponse(FixtureResponse):
    """ A response to an auth request. """

    def __init__(self, *args):
        headers = [("Reply-Text", " ".join(args))]
        super(FixtureAuthResponse, self).__init__(
            self.AUTH_REQUEST, headers=headers)


class FixtureApiResponse(FixtureResponse):
    """ A reply to an ESL command. """

    def __init__(self, *args):
        content = " ".join(args)
        super(FixtureApiResponse, self).__init__(self.API_RESPONSE, content)


class FixtureNotFound(Exception):
    """ Raise when a recording server has no matching fixture. """


class RecordingServerFactory(Factory):
    """ Factory for RecordingServer protocols. """

    protocol = RecordingServer

    def __init__(self, fail_connect=False, uuid=uuid4):
        self.fixtures = []
        self.clients = []

    def add_fixture(self, cmd, response):
        self.fixtures.append((cmd, response))

    def get_response(self, received_cmd):
        for i, (cmd, response) in enumerate(self.fixtures):
            if received_cmd == cmd:
                del self.fixtures[i]
                return response
        raise FixtureNotFound(received_cmd)


class FakeFreeSwitchProtocol(LineReceiver):
    """ A fake connection from FreeSwitch. """

    def __init__(self, call_uuid):
        self.call_uuid = call_uuid
        self.esl_parser = EslParser()
        self.queue = DeferredQueue()
        self.connect_d = Deferred()
        self.disconnect_d = Deferred()
        self.setRawMode()

    def connectionMade(self):
        self.connected = True
        self.connect_d.callback(None)

    def sendPlainEvent(self, name, params=None):
        params = {} if params is None else params
        params['Event-Name'] = name
        data = "\n".join("%s: %s" % (k, v) for k, v in params.items()) + "\n"
        self.sendLine(
            'Content-Length: %d\nContent-Type: text/event-plain\n\n%s' %
            (len(data), data))

    def sendCommandReply(self, params=""):
        self.sendLine('Content-Type: command/reply\nReply-Text: +OK\n%s\n\n' %
                      params)

    def sendChannelHangupCompleteEvent(self, duration):
        """
        Sends a hangup complete event. Duration = duration of call in ms
        """
        hangup_time = int(time.time() * 1000)
        answer_time = int(hangup_time - duration)
        self.sendPlainEvent('Channel_Hangup_Complete', {
            'Caller-Channel-Answered-Time': answer_time,
            'Caller-Channel-Hangup-Time': hangup_time
        })

    def sendDtmfEvent(self, digit):
        self.sendPlainEvent('DTMF', {
            'DTMF-Digit': digit,
        })

    def sendDisconnectEvent(self):
        self.sendLine('Content-Type: text/disconnect-notice\n\n')

    def sendChannelAnswerEvent(self):
        self.sendPlainEvent('Channel_Answer', {
            'Variable-Caller-ID': self.call_uuid
        })

    def rawDataReceived(self, data):
        for cmd in self.esl_parser.parse(data):
            if cmd.cmd_type == "connect":
                self.sendCommandReply(
                    'variable-call-uuid: %s' % self.call_uuid)
            elif cmd.cmd_type == "myevents":
                self.sendCommandReply()
            elif cmd.cmd_type == "sendmsg":
                self.sendCommandReply()
                cmd_name = cmd.params.get('execute-app-name')
                if cmd_name == "speak":
                    self.queue.put(cmd)
                elif cmd_name == "playback":
                    self.queue.put(cmd)
                elif cmd_name == "play_and_get_digits":
                    self.queue.put(cmd)

    def connectionLost(self, reason):
        self.connected = False
        self.disconnect_d.callback(None)


class EslHelper(object):
    """
    Test helper for working with ESL servers.
    """

    implements(IHelper)

    def __init__(self):
        self._recorders = []
        self._clients = []

    def setup(self):
        pass

    @inlineCallbacks
    def cleanup(self):
        for server, factory in self._recorders:
            yield server.stopListening()
            for client in factory.clients:
                yield client.transport.loseConnection()
        for client in self._clients:
            if client.transport and client.transport.connected:
                client.sendDisconnectEvent()
                yield client.transport.loseConnection()
                yield client.disconnect_d

    @proxyable
    @inlineCallbacks
    def mk_server(self, port=1337, fail_connect=False, uuid=uuid4):
        endpoint = TCP4ServerEndpoint(reactor, port)
        factory = RecordingServerFactory(fail_connect=fail_connect, uuid=uuid)
        server = yield endpoint.listen(factory)
        self._recorders.append((server, factory))
        returnValue(factory)

    @proxyable
    @inlineCallbacks
    def mk_client(self, worker, call_uuid="test-uuid"):
        addr = worker.voice_server.getHost()
        client = FakeFreeSwitchProtocol(call_uuid)
        self._clients.append(client)
        factory = ClientFactory.forProtocol(lambda: client)
        yield reactor.connectTCP("127.0.0.1", addr.port, factory)
        yield client.connect_d
        returnValue(client)



# coding: utf-8

"""Tests for vxfreeswitch.voice."""

import logging
import md5
import os

from twisted.internet.defer import inlineCallbacks

from vumi.message import TransportUserMessage
from vumi.tests.helpers import VumiTestCase
from vumi.tests.utils import LogCatcher
from vumi.transports.tests.helpers import TransportHelper

from vxfreeswitch import VoiceServerTransport
from vxfreeswitch.voice import FreeSwitchESLProtocol
from vxfreeswitch.tests.helpers import (
    EslCommand, EslHelper, EslTransport, FixtureApiResponse)


class TestFreeSwitchESLProtocol(VumiTestCase):

    transport_class = VoiceServerTransport

    VOICE_CMD = """
        python -c open("{filename}","w").write("{text}")
    """

    @inlineCallbacks
    def setUp(self):
        self.esl_helper = self.add_helper(EslHelper())
        self.tx_helper = self.add_helper(
            TransportHelper(self.transport_class))
        self.worker = yield self.tx_helper.get_transport({
            'twisted_endpoint': 'tcp:port=0',
            'freeswitch_endpoint': 'tcp:127.0.0.1:1337',
            'originate_parameters': {
                'call_url': '/sofia/gateway/yogisip',
                'exten': '100',
                'cid_name': 'elcid',
                'cid_num': '+1234'
            },
        })
        self.tr = EslTransport()

        self.proto = FreeSwitchESLProtocol(self.worker)
        self.proto.transport = self.tr

        self.voice_cache_folder = self.mktemp()
        os.mkdir(self.voice_cache_folder)

    def send_event(self, params):
        for key, value in params:
            self.proto.dataReceived("%s:%s\n" % (key, value))
        self.proto.dataReceived("\n")

    def send_command_reply(self, response):
        self.send_event([
            ("Content_Type", "command/reply"),
            ("Reply_Text", response),
        ])

    @inlineCallbacks
    def assert_and_reply(self, expected, response):
        cmd = yield self.tr.cmds.get()
        expected_cmd = EslCommand.from_dict(expected)
        self.assertEqual(cmd, expected_cmd)
        self.send_command_reply(response)

    @inlineCallbacks
    def assert_and_reply_tts(self, engine, voice, msg):
        yield self.assert_and_reply({
            "type": "sendmsg", "name": "set",
            "arg": "tts_engine=%s" % engine,
        }, "+OK")
        yield self.assert_and_reply({
            "type": "sendmsg", "name": "set",
            "arg": "tts_voice=%s" % voice,
        }, "+OK")
        yield self.assert_and_reply_playback("say:'%s'" % msg)

    @inlineCallbacks
    def assert_and_reply_playback(self, url):
        yield self.assert_and_reply({
            "type": "sendmsg", "name": "set",
            "arg": "playback_terminators=None",
        }, "+OK")
        yield self.assert_and_reply({
            "type": "sendmsg", "name": "playback",
            "arg": url,
        }, "+OK")

    @inlineCallbacks
    def assert_and_reply_get_digits(self, msg, **kwargs):
        params = {
            'minimum': 1, 'maximum': 1, 'timeout': 3000, 'terminator': "''",
            'tries': 1, 'msg': msg,
        }
        params.update(kwargs)
        yield self.assert_and_reply({
            "type": "sendmsg", "name": "play_and_get_digits",
            "arg": ("%(minimum)d %(maximum)d %(tries)d %(timeout)d "
                    "%(terminator)s %(msg)s silence_stream://1") % params,
        }, "+OK")

    def test_create_tts_command(self):
        self.assertEqual(
            self.proto.create_tts_command("foo", "myfile", "hi!"),
            ("foo", []))
        self.assertEqual(
            self.proto.create_tts_command(
                "foo -f {filename} -t {text}", "myfile", "hi!"),
            ("foo", ["-f", "myfile", "-t", "hi!"]))

    @inlineCallbacks
    def test_create_and_stream_text_as_speech_file_found(self):
        self.proto.uniquecallid = "abc-1234"
        content = "Hello!"
        voice_key = md5.md5(content).hexdigest()
        voice_filename = os.path.join(
            self.voice_cache_folder, "voice-%s.wav" % voice_key)
        with open(voice_filename, "w") as f:
            f.write("Dummy voice file")

        with LogCatcher() as lc:
            d = self.proto.create_and_stream_text_as_speech(
                self.voice_cache_folder, self.VOICE_CMD, "wav", content)
            self.assertEqual(lc.messages(), [
                "[abc-1234] Using cached voice file %r" % (voice_filename,),
                "[abc-1234] Playing back: %r" % (voice_filename,),
            ])

        yield self.assert_and_reply_playback(voice_filename)
        yield d

        with open(voice_filename) as f:
            self.assertEqual(f.read(), "Dummy voice file")

    @inlineCallbacks
    def test_create_and_stream_text_as_speech_file_not_found(self):
        self.proto.uniquecallid = "abc-1234"
        content = "Hello!"
        voice_key = md5.md5(content).hexdigest()
        voice_filename = os.path.join(
            self.voice_cache_folder, "voice-%s.wav" % voice_key)

        with LogCatcher() as lc:
            d = self.proto.create_and_stream_text_as_speech(
                self.voice_cache_folder, self.VOICE_CMD, "wav", content)
            self.assertEqual(lc.messages(), [
                "[abc-1234] Generating voice file %r" % (voice_filename,)
            ])

        yield self.assert_and_reply_playback(voice_filename)
        yield d

        with open(voice_filename) as f:
            self.assertEqual(f.read(), "Hello!")

    @inlineCallbacks
    def test_send_text_as_speech(self):
        d = self.proto.send_text_as_speech(
            "thomas", "his_masters_voice", "hi!")
        yield self.assert_and_reply_tts("thomas", "his_masters_voice", "hi!")
        yield d

    @inlineCallbacks
    def test_output_message(self):
        self.proto.uniquecallid = "abc-1234"
        with LogCatcher() as lc:
            d = self.proto.output_message("Foo!")
            yield self.assert_and_reply_tts("flite", "kal", "Foo!")
            yield d
            self.assertEqual(lc.messages(), [
                "[abc-1234] Playing back: \"say:'Foo!'\"",
            ])

    @inlineCallbacks
    def test_output_stream(self):
        self.proto.uniquecallid = "abc-1234"
        voice_filename = "http://example.com/foo.mp3"
        with LogCatcher() as lc:
            d = self.proto.output_stream(voice_filename)
            self.assertEqual(lc.messages(), [
                "[abc-1234] Playing back: 'http://example.com/foo.mp3'",
            ])
        yield self.assert_and_reply_playback(voice_filename)
        yield d

    def test_unboundEvent(self):
        self.proto.uniquecallid = "abc-1234"
        with LogCatcher() as lc:
            self.proto.unboundEvent({"some": "data"}, "custom_event")
            self.assertEqual(lc.messages(), [
                "[abc-1234] Unbound event 'custom_event'",
            ])

    @inlineCallbacks
    def test_output_stream_barge_in_defaults(self):
        self.proto.output_stream('foo', {'barge_in': True})
        yield self.assert_and_reply_get_digits('foo')

    @inlineCallbacks
    def test_output_stream_barge_in_non_defaults(self):
        self.proto.output_stream('foo', {
            'barge_in': True, 'wait_for': '#', 'tries': 2, 'time_gap': 5000})
        yield self.assert_and_reply_get_digits(
            'foo', minimum=0, maximum=128, tries=2, timeout=5000,
            terminator='#')

    @inlineCallbacks
    def test_send_text_as_speech_quote_escaping(self):
        '''If there are any single quotes in the text that we are converting
        to speech, we should escape those quotes before sending it to
        freeswitch, since we send the text string within single quotes.'''
        d = self.proto.send_text_as_speech(
            "thomas", "his_masters_voice", "text with single quote's")
        yield self.assert_and_reply_tts(
            "thomas", "his_masters_voice", "text with single quote\\'s")
        yield d


class TestVoiceServerTransportInboundCalls(VumiTestCase):

    transport_class = VoiceServerTransport
    transport_type = 'voice'

    @inlineCallbacks
    def setUp(self):
        self.tx_helper = self.add_helper(TransportHelper(self.transport_class))
        self.esl_helper = self.add_helper(EslHelper())
        self.worker = yield self.tx_helper.get_transport({
            'twisted_endpoint': 'tcp:port=0',
            'freeswitch_endpoint': 'tcp:127.0.0.1:1337',
            'originate_parameters': {
                'call_url': '/sofia/gateway/yogisip',
                'exten': '100',
                'cid_name': 'elcid',
                'cid_num': '+1234'
            },
        })
        self.client = yield self.esl_helper.mk_client(self.worker)

    def assert_get_digits_command(self, cmd, msg, **kwargs):
        params = {
            'minimum': 1, 'maximum': 1, 'timeout': 3000, 'terminator': "''",
            'tries': 1, 'msg': msg,
        }
        params.update(kwargs)
        self.assertEqual(cmd, EslCommand.from_dict({
            'type': 'sendmsg', 'name': 'play_and_get_digits',
            "arg": ("%(minimum)d %(maximum)d %(tries)d %(timeout)d "
                    "%(terminator)s %(msg)s silence_stream://1") % params,
        }))

    @inlineCallbacks
    def test_client_register(self):
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], None)
        self.assertEqual(msg['session_event'],
                         TransportUserMessage.SESSION_NEW)

    @inlineCallbacks
    def test_client_deregister(self):
        # wait for registration message
        yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()
        self.client.sendDisconnectEvent()
        self.client.transport.loseConnection()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], None)
        self.assertEqual(msg['session_event'],
                         TransportUserMessage.SESSION_CLOSE)

    @inlineCallbacks
    def test_client_hangup_and_disconnect(self):
        duration = 20
        yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()
        self.client.sendChannelHangupCompleteEvent(duration)
        self.client.sendDisconnectEvent()
        self.client.transport.loseConnection()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], None)
        self.assertEqual(msg['session_event'],
                         TransportUserMessage.SESSION_CLOSE)
        self.assertEqual(
            msg['helper_metadata']['voice']['call_duration'], duration)

    @inlineCallbacks
    def test_client_hangup_invalid_freeswitch_duration(self):
        yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()
        self.client.sendPlainEvent('Channel_Hangup_Complete', {
            'Caller-Channel-Answered-Time': 'foo',
            'Caller-Channel-Hangup-Time': 'bar'
        })
        self.client.transport.loseConnection()
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], None)
        self.assertEqual(msg['session_event'],
                         TransportUserMessage.SESSION_CLOSE)
        self.assertEqual(msg['helper_metadata'].get('voice'), None)

    @inlineCallbacks
    def test_simplemessage(self):
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        msg = yield self.tx_helper.make_dispatch_reply(reg, "voice test")

        cmd = yield self.client.queue.get()
        self.assertEqual(cmd, EslCommand.from_dict({
            'type': 'sendmsg', 'name': 'playback', 'arg': "say:'voice test . '",
        }))

        [ack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(ack['user_message_id'], msg['message_id'])
        self.assertEqual(ack['sent_message_id'], msg['message_id'])

    @inlineCallbacks
    def test_simpledigitcapture(self):
        yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()
        self.client.sendDtmfEvent('5')
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], '5')

    @inlineCallbacks
    def test_multidigitcapture(self):
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()

        yield self.tx_helper.make_dispatch_reply(
            reg, 'voice test', helper_metadata={'voice': {'wait_for': '#'}})

        cmd = yield self.client.queue.get()
        self.assertEqual(cmd, EslCommand.from_dict({
            'type': 'sendmsg', 'name': 'playback', 'arg': "say:'voice test . '",
        }))

        self.client.sendDtmfEvent('5')
        self.client.sendDtmfEvent('7')
        self.client.sendDtmfEvent('2')
        self.client.sendDtmfEvent('#')
        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], '572')

    @inlineCallbacks
    def test_speech_url_string(self):
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()

        msg = yield self.tx_helper.make_dispatch_reply(
            reg, 'speech url test', helper_metadata={
                'voice': {
                    'speech_url': 'http://example.com/speech_url_test.ogg'
                }
            })

        cmd = yield self.client.queue.get()
        self.assertEqual(cmd, EslCommand.from_dict({
            'type': 'sendmsg', 'name': 'playback',
            'arg': 'http://example.com/speech_url_test.ogg',
        }))

        [ack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(ack['user_message_id'], msg['message_id'])
        self.assertEqual(ack['sent_message_id'], msg['message_id'])

    @inlineCallbacks
    def test_speech_url_list(self):
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()
        urls = [
            'http://example.com/speech_url_test1.ogg',
            'http://example.com/speech_url_test2.ogg'
        ]

        msg = yield self.tx_helper.make_dispatch_reply(
            reg, 'speech url test', helper_metadata={
                'voice': {
                    'speech_url': urls,
                }
            })

        cmd = yield self.client.queue.get()
        urllist = 'file_string://%s' % '!'.join(urls)
        self.assertEqual(cmd, EslCommand.from_dict({
            'type': 'sendmsg', 'name': 'playback',
            'arg': urllist,
        }))

        [ack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(ack['user_message_id'], msg['message_id'])
        self.assertEqual(ack['sent_message_id'], msg['message_id'])
        self.assertEqual(ack['event_type'], 'ack')

    @inlineCallbacks
    def test_speech_url_invalid_url(self):
        url = 7
        with LogCatcher(log_level=logging.WARN) as lc:
            [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
            self.tx_helper.clear_dispatched_inbound()

            msg = yield self.tx_helper.make_dispatch_reply(
                reg, 'speech url test', helper_metadata={
                    'voice': {
                        'speech_url': url
                    }
                })
        [warn_log] = lc.messages()
        self.assertEqual(warn_log, "Invalid URL %r" % url)

        [nack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(nack['user_message_id'], msg['message_id'])
        self.assertEqual(nack['event_type'], 'nack')
        self.assertEqual(nack['nack_reason'], 'Invalid URL %r' % url)

    @inlineCallbacks
    def test_speech_invalid_url_list(self):
        valid_url = u'http://example.com/speech_url_test1.ogg'
        invalid_url1 = 7
        invalid_url2 = 8
        with LogCatcher(log_level=logging.WARN) as lc:
            [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
            self.tx_helper.clear_dispatched_inbound()

            msg = yield self.tx_helper.make_dispatch_reply(
                reg, 'speech url test', helper_metadata={
                    'voice': {
                        'speech_url': [
                            invalid_url1,
                            valid_url,
                            invalid_url2,
                        ]
                    }
                })

        [log] = lc.messages()
        self.assertEqual(log, 'Invalid URL list %r' % (
            [invalid_url1, valid_url, invalid_url2], ))
        [nack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(nack['user_message_id'], msg['message_id'])
        self.assertEqual(nack['event_type'], 'nack')
        self.assertEqual(nack['nack_reason'], 'Invalid URL list %r' % (
            [invalid_url1, valid_url, invalid_url2], ))

    @inlineCallbacks
    def test_reply_to_client_that_has_hung_up(self):
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()
        [client] = self.worker._clients.values()
        self.worker.deregister_client(client)

        msg = yield self.tx_helper.make_dispatch_reply(
            reg, 'speech url test', helper_metadata={
                'voice': {
                    'speech_url': 'http://example.com/speech_url_test.ogg'
                }
            })

        [nack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(nack['user_message_id'], msg['message_id'])
        self.assertEqual(nack['nack_reason'],
                         "Client u'test-uuid' no longer connected")

    @inlineCallbacks
    def test_barge_in_defaults(self):
        '''Barge ins should use the play_and_get_digits command with certain
        default values.'''
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()

        yield self.tx_helper.make_dispatch_reply(
            reg, 'barge in test', helper_metadata={
                'voice': {
                    'barge_in': True,
                },
            })

        cmd = yield self.client.queue.get()
        self.assert_get_digits_command(cmd, "say:'barge in test . '")

    @inlineCallbacks
    def test_barge_in_non_defaults(self):
        '''When the correct fields are specified, these should override the
        defaults.'''
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()

        yield self.tx_helper.make_dispatch_reply(
            reg, 'barge in test', helper_metadata={
                'voice': {
                    'barge_in': True,
                    'wait_for': '#',
                    'tries': 2,
                    'time_gap': 5000,
                },
            })

        cmd = yield self.client.queue.get()
        self.assert_get_digits_command(
            cmd, "say:'barge in test . '", minimum=0, maximum=128,
            terminator='#', tries=2, timeout=5000)

    @inlineCallbacks
    def test_barge_in_collecting_digits(self):
        '''If we send a barge in message, we should collect the digits that
        the client has sent us.'''
        [reg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.tx_helper.clear_dispatched_inbound()

        yield self.tx_helper.make_dispatch_reply(
            reg, 'barge in test', helper_metadata={
                'voice': {
                    'barge_in': True,
                    'wait_for': '#',
                },
            })

        self.client.sendDtmfEvent('5')
        self.client.sendDtmfEvent('6')
        self.client.sendDtmfEvent('#')

        [msg] = yield self.tx_helper.wait_for_dispatched_inbound(1)
        self.assertEqual(msg['content'], '56')


class TestVoiceServerTransportOutboundCalls(VumiTestCase):

    transport_class = VoiceServerTransport

    @inlineCallbacks
    def setUp(self):
        self.tx_helper = self.add_helper(TransportHelper(self.transport_class))
        self.esl_helper = self.add_helper(EslHelper())
        self.worker = yield self.tx_helper.get_transport({
            'twisted_endpoint': 'tcp:port=0',
            'freeswitch_endpoint': 'tcp:127.0.0.1:port=1337',
            'originate_parameters': {
                'call_url': '/sofia/gateway/yogisip',
                'exten': '100',
                'cid_name': 'elcid',
                'cid_num': '+1234'
            },
        })

    @inlineCallbacks
    def test_create_call(self):
        factory = yield self.esl_helper.mk_server()
        factory.add_fixture(
            EslCommand("api originate /sofia/gateway/yogisip"
                       " 100 XML default elcid +1234 60"),
            FixtureApiResponse("+OK uuid-1234"))

        msg = self.tx_helper.make_outbound(
            'foobar', '12345', '54321', session_event='new')

        client = yield self.esl_helper.mk_client(self.worker, 'uuid-1234')

        with LogCatcher(log_level=logging.WARN) as lc:
            yield self.tx_helper.dispatch_outbound(msg)
        self.assertEqual(lc.messages(), [])

        events = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(events, [])

        client.sendChannelAnswerEvent()

        cmd = yield client.queue.get()
        self.assertEqual(cmd, EslCommand.from_dict({
            'type': 'sendmsg', 'name': 'playback', 'arg': "say:'foobar . '",
        }))

        [ack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assertEqual(ack['event_type'], 'ack')
        self.assertEqual(ack['sent_message_id'], msg['message_id'])

    @inlineCallbacks
    def test_connect_error(self):
        factory = yield self.esl_helper.mk_server(
            fail_connect=True, uuid=lambda: 'uuid-1234')
        factory.add_fixture(
            EslCommand("api originate /sofia/gateway/yogisip"
                       " 100 XML default elcid +1234 60"),
            FixtureApiResponse("+ERROR Bad horse."))

        msg = self.tx_helper.make_outbound(
            'foobar', '12345', '54321', session_event='new')
        with LogCatcher(message='Error connecting') as lc:
            yield self.tx_helper.dispatch_outbound(msg)
        self.assertEqual(lc.messages(), [
            "Error connecting to client u'54321':"
            " +ERROR Bad horse.",
        ])
        [nack] = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(nack['user_message_id'], msg['message_id'])
        self.assertEqual(nack['nack_reason'],
                         "Could not make call to client u'54321'")

    @inlineCallbacks
    def test_client_disconnect_without_answer(self):
        factory = yield self.esl_helper.mk_server()
        factory.add_fixture(
            EslCommand("api originate /sofia/gateway/yogisip"
                       " 100 XML default elcid +1234 60"),
            FixtureApiResponse("+OK uuid-1234"))

        msg = self.tx_helper.make_outbound(
            'foobar', '12345', '54321', session_event='new')

        client = yield self.esl_helper.mk_client(self.worker, 'uuid-1234')

        with LogCatcher(log_level=logging.WARN) as lc:
            yield self.tx_helper.dispatch_outbound(msg)

        self.assertEqual(lc.messages(), [])

        events = yield self.tx_helper.get_dispatched_events()
        self.assertEqual(events, [])

        client.sendDisconnectEvent()

        [nack] = yield self.tx_helper.wait_for_dispatched_events(1)
        self.assertEqual(nack['event_type'], 'nack')
        self.assertEqual(nack['nack_reason'], 'Unanswered Call')
        self.assertEqual(nack['user_message_id'], msg['message_id'])



""" Tests for vxfreeswitch.client. """

from zope.interface import implementer

from twisted.internet.interfaces import IStreamClientEndpoint
from twisted.internet.defer import inlineCallbacks, Deferred, fail, succeed
from twisted.internet.protocol import ClientFactory
from twisted.test.proto_helpers import StringTransportWithDisconnection
from twisted.python.failure import Failure
from twisted.trial.unittest import TestCase

from eventsocket import EventError

from vxfreeswitch.client import (
    FreeSwitchClientProtocol, FreeSwitchClientFactory,
    FreeSwitchClient, FreeSwitchClientReply, FreeSwitchClientError)

from vxfreeswitch.tests.helpers import FixtureApiResponse, FixtureReply


def connect_transport(protocol, factory=None):
    """ Connect a StringTransport to a client protocol. """
    if factory is None:
        factory = ClientFactory()
    transport = StringTransportWithDisconnection()
    protocol.makeConnection(transport)
    transport.protocol = protocol
    protocol.factory = factory
    return transport


@implementer(IStreamClientEndpoint)
class StringClientEndpoint(object):
    """ Client endpoint that connects to a StringTransport. """
    transport = None

    def connect(self, factory):
        try:
            protocol = factory.buildProtocol("dummy-address")
            self.transport = connect_transport(protocol, factory)
        except:
            return fail()
        return succeed(protocol)


class TestFreeSwitchClientProtocol(TestCase):
    def test_connected(self):
        p = FreeSwitchClientProtocol(auth=None)
        self.assertTrue(isinstance(p._connected, Deferred))
        self.assertEqual(p._connected.called, False)
        connect_transport(p)
        self.assertEqual(p._connected.called, True)
        self.assertEqual(p._connected.result, p)

    def test_auth(self):
        p = FreeSwitchClientProtocol(auth="pw-12345")
        tr = connect_transport(p)
        self.assertEqual(tr.value(), "auth pw-12345\n\n")

    def test_no_auth(self):
        p = FreeSwitchClientProtocol(auth=None)
        tr = connect_transport(p)
        self.assertEqual(tr.value(), "")


class TestFreeSwitchClientFactory(TestCase):
    def test_subclasses_client_factory(self):
        f = FreeSwitchClientFactory()
        self.assertTrue(isinstance(f, ClientFactory))

    def test_protocol(self):
        f = FreeSwitchClientFactory()
        p = f.protocol()
        self.assertTrue(isinstance(p, FreeSwitchClientProtocol))

    def test_default_noisy(self):
        f = FreeSwitchClientFactory()
        self.assertEqual(f.noisy, False)

    def test_set_noisy(self):
        f = FreeSwitchClientFactory(noisy=True)
        self.assertEqual(f.noisy, True)

    def test_no_auth(self):
        f = FreeSwitchClientFactory()
        p = f.protocol()
        tr = connect_transport(p)
        self.assertEqual(tr.value(), "")

    def test_auth(self):
        f = FreeSwitchClientFactory(auth="pw-1234")
        p = f.protocol()
        tr = connect_transport(p)
        self.assertEqual(tr.value(), "auth pw-1234\n\n")


class TestFreeSwitchClientError(TestCase):
    def test_subclasses_exception(self):
        err = FreeSwitchClientError("foo")
        self.assertTrue(isinstance(err, Exception))

    def test_str(self):
        err = FreeSwitchClientError("reason")
        self.assertEqual(str(err), "reason")


class TestFreeSwitchClientReply(TestCase):
    def test_args(self):
        reply = FreeSwitchClientReply("a", "b")
        self.assertEqual(reply.args, ("a", "b"))

    def test_repr(self):
        self.assertEqual(
            repr(FreeSwitchClientReply("a", "c")),
            "<FreeSwitchClientReply args=('a', 'c')>")

    def test_equal(self):
        self.assertEqual(
            FreeSwitchClientReply("a", "b"),
            FreeSwitchClientReply("a", "b"))

    def test_not_equal(self):
        self.assertNotEqual(
            FreeSwitchClientReply("a", "b"),
            FreeSwitchClientReply("a", "c"))

    def test_not_equal_other_object(self):
        self.assertNotEqual(
            FreeSwitchClientReply("a", "b"),
            object())


class TestFreeSwitchClient(TestCase):
    def mk_client(self, endpoint=None, auth=None):
        return FreeSwitchClient(endpoint=endpoint, auth=auth)

    def test_fallback_error_handler_client_error(self):
        client = self.mk_client()
        failure = Failure(FreeSwitchClientError("reason"))
        self.assertEqual(
            client.fallback_error_handler(failure), failure)

    def test_fallback_error_handler_other_error(self):
        client = self.mk_client()
        failure = Failure(Exception("reason"))
        err = self.failUnlessRaises(
            FreeSwitchClientError,
            client.fallback_error_handler, failure)
        self.assertEqual(str(err), "reason")

    def test_event_error_handler_event_error_has_reply(self):
        client = self.mk_client()
        failure = Failure(EventError({"Reply_Text": "+ERROR eep"}))
        err = self.failUnlessRaises(
            FreeSwitchClientError,
            client.event_error_handler, failure)
        self.assertEqual(str(err), "+ERROR eep")

    def test_event_error_handler_event_error_no_reply(self):
        client = self.mk_client()
        failure = Failure(EventError({"Not_Reply": "foo"}))
        err = self.failUnlessRaises(
            FreeSwitchClientError,
            client.event_error_handler, failure)
        self.assertEqual(str(err), "{'Not_Reply': 'foo'}")

    def test_event_error_handler_other_error(self):
        client = self.mk_client()
        failure = Failure(Exception("reason"))
        self.assertEqual(
            client.event_error_handler(failure), failure)

    def test_request_callback_with_reply(self):
        client = self.mk_client()
        self.assertEqual(
            client.request_callback({'Reply_Text': 'a b'}),
            FreeSwitchClientReply('a', 'b'))

    def test_request_callback_without_reply(self):
        client = self.mk_client()
        self.assertEqual(
            client.request_callback({}),
            FreeSwitchClientReply())

    def test_api_request_callback_with_okay_response(self):
        client = self.mk_client()
        self.assertEqual(
            client.api_request_callback({
                'data': {
                    'rawresponse': '+OK meep moop'
                }
            }),
            FreeSwitchClientReply('+OK', 'meep', 'moop'))

    def test_api_request_callback_with_error_response(self):
        client = self.mk_client()
        err = self.failUnlessRaises(
            FreeSwitchClientError,
            client.api_request_callback, {
                'data': {
                    'rawresponse': '+ERROR meep moop'
                }
            })
        self.assertEqual(str(err), "+ERROR meep moop")

    def test_api_request_callback_without_data(self):
        client = self.mk_client()
        err = self.failUnlessRaises(
            FreeSwitchClientError,
            client.api_request_callback, {
                'foo': 'bar',
            })
        self.assertEqual(str(err), "{'foo': 'bar'}")

    def test_api_request_callback_without_rawresponse(self):
        client = self.mk_client()
        err = self.failUnlessRaises(
            FreeSwitchClientError,
            client.api_request_callback, {
                'data': {}
            })
        self.assertEqual(str(err), "{'data': {}}")

    @inlineCallbacks
    def test_with_connection(self):
        endpoint = StringClientEndpoint()
        client = self.mk_client(endpoint=endpoint)
        f_called = Deferred()

        def f(conn):
            wait = Deferred()
            f_called.callback((wait, conn))
            return wait

        d = client.with_connection(f)
        self.assertEqual(endpoint.transport.connected, True)
        self.assertEqual(endpoint.transport.value(), "")
        self.assertTrue(isinstance(d.result, Deferred))

        f_wait, f_conn = yield f_called
        self.assertTrue(isinstance(f_conn, FreeSwitchClientProtocol))
        self.assertTrue(isinstance(d.result, Deferred))
        self.assertEqual(f_wait.called, False)

        f_wait.callback({'foo': 'bar'})
        reply = yield d

        self.assertEqual(reply, {'foo': 'bar'})
        self.assertEqual(endpoint.transport.value(), "")
        self.assertEqual(endpoint.transport.connected, False)

    @inlineCallbacks
    def test_api(self):
        endpoint = StringClientEndpoint()
        client = self.mk_client(endpoint=endpoint)

        d = client.api("foo")
        self.assertEqual(endpoint.transport.value(), "api foo\n\n")
        self.assertEqual(endpoint.transport.connected, True)
        endpoint.transport.protocol.dataReceived(
            FixtureApiResponse("+OK moo").to_bytes())
        result = yield d

        self.assertEqual(result, FreeSwitchClientReply("+OK", "moo"))
        self.assertEqual(endpoint.transport.value(), "api foo\n\n")
        self.assertEqual(endpoint.transport.connected, False)

    @inlineCallbacks
    def test_auth(self):
        endpoint = StringClientEndpoint()
        client = self.mk_client(endpoint=endpoint, auth="kenny")
        f_called = Deferred()

        def f(conn):
            wait = Deferred()
            f_called.callback((wait, conn))
            return wait

        d = client.with_connection(f)
        self.assertEqual(endpoint.transport.value(), "auth kenny\n\n")
        self.assertEqual(endpoint.transport.connected, True)
        self.assertEqual(f_called.called, False)
        self.assertTrue(isinstance(d.result, Deferred))

        endpoint.transport.protocol.dataReceived(
            FixtureReply("+OK").to_bytes())

        f_wait, f_conn = yield f_called
        self.assertTrue(isinstance(f_conn, FreeSwitchClientProtocol))
        self.assertEqual(f_wait.called, False)

        f_wait.callback({"foo": "bar"})
        reply = yield d

        self.assertEqual(reply, {"foo": "bar"})
        self.assertEqual(endpoint.transport.value(), "auth kenny\n\n")
        self.assertEqual(endpoint.transport.connected, False)



# -*- test-case-name: vxfreeswitch.tests.test_voice -*-

"""A Vumi transport for connecting to SIP gateways via FreeSWITCH's ESL
interfaces.

If outbound messages provide a URL to a pre-recorded audio message,
that is played via SIP. Otherwise sound is generated using a
text-to-speech engine.

DTMF digits dialed on the phone are collected by the transport and
sent as inbound messages. Digits may either be sent through individually
(the default) or collected until a specified character is pressed.
"""

import logging
import md5
import os

from twisted.internet.protocol import ServerFactory
from twisted.internet.defer import (
    inlineCallbacks, returnValue, Deferred, gatherResults)
from twisted.internet.utils import getProcessOutput

from eventsocket import EventProtocol

from confmodel.errors import ConfigError
from confmodel.fields import ConfigText, ConfigDict

from vumi.transports import Transport
from vumi.message import TransportUserMessage
from vumi.config import ConfigClientEndpoint, ConfigServerEndpoint
from vumi.errors import VumiError

from vxfreeswitch.originate import (
    OriginateFormatter, OriginateMissingParameter)
from vxfreeswitch.client import FreeSwitchClient, FreeSwitchClientError


class VoiceError(VumiError):
    """Raised when errors occur while processing voice messages."""


class FreeSwitchESLProtocol(EventProtocol):

    def __init__(self, vumi_transport):
        EventProtocol.__init__(self)
        self.vumi_transport = vumi_transport
        self.request_hang_up = False
        self.current_input = ''
        self.input_type = None
        self.uniquecallid = None

    def unknownContentType(self, content_type, ctx):
        self.vumi_transport.log.debug(
            "[eventsocket] unknown Content-Type: %s" % content_type)

    def unboundEvent(self, ctx, evname):
        self.vumi_transport.log.debug(
            "[eventsocket] unbound Event: %s" % evname)

    @inlineCallbacks
    def connectionMade(self):
        yield self.connect().addCallback(self.on_connect)
        yield self.myevents()
        yield self.answer()
        yield self.vumi_transport.register_client(self)

    def log(self, msg, level=logging.INFO):
        self.vumi_transport.log.msg(
            '[%s] %s' % (self.uniquecallid, msg), logLevel=level)

    def on_connect(self, ctx):
        self.uniquecallid = ctx.variable_call_uuid

    def onDtmf(self, ev):
        if self.input_type is None:
            return self.vumi_transport.handle_input(self, ev.DTMF_Digit)
        else:
            if ev.DTMF_Digit == self.input_type:
                ret_value = self.current_input
                self.current_input = ''
                return self.vumi_transport.handle_input(self, ret_value)
            else:
                self.current_input += ev.DTMF_Digit

    def create_tts_command(self, command_template, filename, message):
        params = {"filename": filename, "text": message}
        args = command_template.strip().split()
        cmd, args = args[0], args[1:]
        args = [arg.format(**params) for arg in args]
        return cmd, args

    @inlineCallbacks
    def create_and_stream_text_as_speech(
            self, folder, command, ext, message, settings={}):
        key = md5.md5(message).hexdigest()
        filename = os.path.join(folder, "voice-%s.%s" % (key, ext))
        if not os.path.exists(filename):
            self.log("Generating voice file %r" % (filename,))
            cmd, args = self.create_tts_command(command, filename, message)
            yield getProcessOutput(cmd, args=args)
        else:
            self.log("Using cached voice file %r" % (filename,))

        yield self.output_stream(filename, settings)

    @inlineCallbacks
    def send_text_as_speech(self, engine, voice, message, settings={}):
        yield self.set("tts_engine=" + engine)
        yield self.set("tts_voice=" + voice)
        # 'say:' is misleading here, it functions more like 'speak' in this
        # context.
        yield self.output_stream(
            "say:'%s'" % message.replace("'", "\\'"), settings)

    @inlineCallbacks
    def stream_text_as_speech(self, message, settings):
        finalmessage = message.replace("\n", " . ")
        cfg = self.vumi_transport.config
        if cfg.tts_type == "local":
            yield self.create_and_stream_text_as_speech(
                cfg.tts_local_cache, cfg.tts_local_command,
                cfg.tts_local_ext, finalmessage, settings)
        elif cfg.tts_type == "freeswitch":
            yield self.send_text_as_speech(
                cfg.tts_fs_engine, cfg.tts_fs_voice, finalmessage, settings)
        else:
            raise VoiceError("Unknown tts_type %r" % (
                cfg.tts_type,))

    def get_address(self):
        return self.uniquecallid

    def output_message(self, text, settings={}):
        return self.stream_text_as_speech(text, settings=settings)

    def output_stream(self, message, settings={}):
        self.log("Playing back: %r" % (message,))
        if settings.get('barge_in'):
            terminator = settings.get('wait_for')
            if terminator is None:
                # We just want to get 1 digit
                minimum, maximum = 1, 1
                # We don't want a terminating character, but we have to supply
                # one, so we supply a blank one
                terminator = "''"
            else:
                # 128 is the maximum amount of digits that freeswitch can
                # collect
                minimum, maximum = 0, 128
            tries = settings.get('tries', 1)
            timeout = settings.get('time_gap', 3000)
            # We have to have an invalid response message, so we set it to
            # 1ms of silence
            invalid_message = 'silence_stream://1'
            return self.execute('play_and_get_digits', ' '.join([
                str(minimum), str(maximum), str(tries), str(timeout),
                str(terminator), message, invalid_message]))
        else:
            return self.playback(message)

    def set_input_type(self, input_type):
        self.input_type = input_type

    def close_call(self):
        self.request_hang_up = True

    @inlineCallbacks
    def onChannelExecuteComplete(self, ev):
        self.log("execute complete: %s" % ev.Application)
        if self.request_hang_up:
            yield self.hangup()

    def onChannelHangupComplete(self, ev):
        self.log("Channel HangUp")
        try:
            answered_time = int(ev.get('Caller_Channel_Answered_Time'))
            hangup_time = int(ev.get('Caller_Channel_Hangup_Time'))
            duration = hangup_time - answered_time
        except (TypeError, ValueError):
            self.log(
                "Unable to get call duration for %r" % self.get_address(),
                level=logging.WARNING)
            duration = None
        self.vumi_transport.deregister_client(self, duration)

    def onDisconnect(self, ev):
        self.log("Channel disconnect received")
        self.vumi_transport.deregister_client(self)

    def onChannelAnswer(self, ev):
        self.log("Channel answered")
        self.vumi_transport.client_answered(self)

    def unboundEvent(self, evdata, evname):
        self.log("Unbound event %r" % (evname,), level=logging.DEBUG)


class FreeSwitchESLFactory(ServerFactory):
    """ FreeSwitch ESL server factory. """
    def __init__(self, vumi_transport):
        self.vumi_transport = vumi_transport

    def protocol(self):
        return FreeSwitchESLProtocol(self.vumi_transport)


class VoiceServerTransportConfig(Transport.CONFIG_CLASS):
    """
    Configuration parameters for the voice transport
    """

    to_addr = ConfigText(
        "The ``to_addr`` to use for inbound messages.",
        default="freeswitchvoice", static=True)

    tts_type = ConfigText(
        "Either 'freeswitch' or 'local' to specify where TTS is executed.",
        default="freeswitch", static=True)

    tts_fs_engine = ConfigText(
        "Specify Freeswitch TTS engine to use (only affects tts_type"
        " 'freeswitch').",
        default="flite", static=True)

    tts_fs_voice = ConfigText(
        "Specify Freeswitch TTS voice to use (only affects tts_type"
        " 'freeswitch').",
        default="kal", static=True)

    tts_local_command = ConfigText(
        "Specify command template to use for generating voice files (only"
        " affects tts_type 'local'). E.g. 'flite -o {filename} -t {text}'."
        " Command parameters are split on whitespace (no shell-like escape"
        " processing is performed on the command).",
        default=None, static=True)

    tts_local_cache = ConfigText(
        "Specify folder to cache voice files (only affects tts_type"
        " 'local').", default=".", static=True)

    tts_local_ext = ConfigText(
        "Specify the file extension used for cached voice files (only affects"
        " tts_type 'local').", default="wav", static=True)

    twisted_endpoint = ConfigServerEndpoint(
        "The endpoint the voice transport will listen on (and that Freeswitch"
        " will connect to).",
        required=True, default="tcp:port=8084", static=True)

    freeswitch_endpoint = ConfigClientEndpoint(
        "The endpoint the voice transport will send originate commands"
        "to (and that Freeswitch listens on).",
        default=None, static=True)

    freeswitch_auth = ConfigText(
        "Password for connecting to the Freeswitch endpoint."
        " None means no authentication credentials are offered.",
        default=None, static=True)

    originate_parameters = ConfigDict(
        "The parameters to pass to the originate command when initiating"
        " outbound calls. This dictionary of parameters is passed to the"
        " originate call template:\n\n"
        "  %(template)r\n\n"
        "All call parameters are required but the following defaults are"
        " supplied:\n\n"
        "  %(defaults)r" % {
            'template': OriginateFormatter.PROTO_TEMPLATE,
            'defaults': OriginateFormatter.DEFAULT_PARAMS,
        },
        default=None, static=True)

    @property
    def supports_outbound(self):
        return self.freeswitch_endpoint is not None

    def post_validate(self):
        super(VoiceServerTransportConfig, self).post_validate()
        required_outbound = (
            self.freeswitch_endpoint is not None,
            self.originate_parameters is not None)
        if self.supports_outbound and not all(required_outbound):
            raise ConfigError(
                "If any outbound message parameters are supplied"
                " (freeswitch_endpoint or originate_params), all must be"
                " given.")
        if self.originate_parameters is not None:
            try:
                OriginateFormatter(**self.originate_parameters)
            except OriginateMissingParameter as err:
                raise ConfigError(str(err))


class VoiceServerTransport(Transport):
    """
    Transport for Freeswitch Voice Service.

    Voice transports may receive additional hints for how to handle
    outbound messages the ``voice`` section of ``helper_metadata``.
    The ``voice`` section may contain the following keys:

    * ``speech_url``: An HTTP URL from which a custom sound file to
      use for this message. If absent or ``None`` a text-to-speech
      engine will be used to generate suitable sound to play.
      Sound formats supported are: ``.wav``, ``.ogg`` and ``.mp3``.
      The format will be determined by the ``Content-Type`` returned
      by the URL, or by the file extension if the ``Content-Type``
      header is absent. The preferred format is ``.ogg``.

    * ``wait_for``: Gather response characters until the given
      DTMF character is encountered. Commonly either ``#`` or ``*``.
      If absent or ``None``, an inbound message is sent as soon as
      a single DTMF character arrives.

      If no input is seen for some time (configurable in the
      transport config) the voice transport will timeout the wait
      and send the characters entered so far.

    Example ``helper_metadata``::

      "helper_metadata": {
          "voice": {
              "speech_url": "http://www.example.com/voice/ab34f611cdee.ogg",
              "wait_for": "#",
          },
      }
    """

    CONFIG_CLASS = VoiceServerTransportConfig

    @inlineCallbacks
    def setup_transport(self):
        self._clients = {}
        self._originated_calls = {}
        self._unanswered_channels = {}

        self.config = self.get_static_config()
        self._to_addr = self.config.to_addr
        self._transport_type = "voice"

        if self.config.supports_outbound:
            self.voice_client = FreeSwitchClient(
                self.config.freeswitch_endpoint, self.config.freeswitch_auth)
            self.originate_formatter = OriginateFormatter(
                **self.config.originate_parameters)
        else:
            self.voice_client = None
            self.originate_formatter = None

        self.voice_server = yield self.config.twisted_endpoint.listen(
            FreeSwitchESLFactory(self))

    @inlineCallbacks
    def teardown_transport(self):
        if hasattr(self, 'voice_server'):
            # We need to wait for all the client connections to be closed (and
            # their deregistration messages sent) before tearing down the rest
            # of the transport.
            self.log.info("Shutting down %d clients." % len(self._clients))
            self.voice_server.loseConnection()
            yield gatherResults([
                client.registration_d for client in self._clients.values()])

    @inlineCallbacks
    def register_client(self, client):
        # We add our own Deferred to the client here because we only want to
        # fire it after we're finished with our own deregistration process.
        client.registration_d = Deferred()
        client_addr = client.get_address()
        self.log.info("Registering client connected from %r" % (client_addr,))
        self._clients[client_addr] = client
        originated_msg = self._originated_calls.pop(client_addr, None)
        if originated_msg is not None:
            yield self.send_outbound_message(client, originated_msg)
        else:
            self.send_inbound_message(
                client, None, TransportUserMessage.SESSION_NEW)
        self.log.info("Registration complete.")

    def deregister_client(self, client, duration=None):
        client_addr = client.get_address()

        # If originated call has not yet been answered
        if self._unanswered_channels.get(client_addr):
            d = self._unanswered_channels[client_addr]
            d.errback(FreeSwitchClientError('Call is unanswered'))

        if client_addr not in self._clients:
            return
        self.log.info("Deregistering client connected from %r" % (client_addr,))
        del self._clients[client_addr]

        self.send_inbound_message(
            client, None, TransportUserMessage.SESSION_CLOSE, duration)
        client.registration_d.callback(None)
        self.log.info("Deregistration complete.")

    def handle_input(self, client, text):
        self.send_inbound_message(
            client, text, TransportUserMessage.SESSION_RESUME)

    def send_inbound_message(self, client, text, session_event, duration=None):
        helper_metadata = {}
        if duration:
            if not helper_metadata.get('voice'):
                helper_metadata['voice'] = voice = {}
            voice['call_duration'] = duration

        self.publish_message(
            from_addr=client.get_address(),
            to_addr=self._to_addr,
            session_event=session_event,
            content=text,
            transport_name=self.transport_name,
            transport_type=self._transport_type,
            helper_metadata=helper_metadata,
        )

    @inlineCallbacks
    def log_and_nack(self, message, error):
        self.log.warning(error)
        yield self.publish_nack(
            message["message_id"], reason=error)

    @inlineCallbacks
    def send_outbound_message(self, client, message):
        content = message['content']
        if content is None:
            content = u''
        content = u"\n".join(content.splitlines())
        content = content.encode('utf-8')

        voicemeta = get_in(message, 'helper_metadata', 'voice', default={})
        client.set_input_type(voicemeta.get('wait_for', None))
        overrideURL = voicemeta.get('speech_url', None)

        # Wait if call isn't answered
        if self._unanswered_channels.get(client.get_address()):
            try:
                yield self._unanswered_channels.pop(client.get_address())
            except FreeSwitchClientError:
                yield self.publish_nack(
                    message['message_id'], 'Unanswered Call')
                returnValue(None)

        if overrideURL is None:
            yield client.output_message("%s\n" % content, voicemeta)
        elif isinstance(overrideURL, basestring):
            yield client.output_stream(overrideURL, voicemeta)
        elif isinstance(overrideURL, list):
            try:
                urllist = 'file_string://%s' % '!'.join(overrideURL)
                yield client.output_stream(urllist, voicemeta)
            except TypeError:
                error = "Invalid URL list %r" % overrideURL
                yield self.log_and_nack(message, error)
                return
        else:
            error = "Invalid URL %r" % overrideURL
            yield self.log_and_nack(message, error)
            return

        if message['session_event'] == TransportUserMessage.SESSION_CLOSE:
            client.close_call()

        yield self.publish_ack(message["message_id"], message["message_id"])

    def client_answered(self, client):
        """Function that is called when the ChannelAnswer event is received.
        Fires the deferred related to the outbound call"""
        d = self._unanswered_channels.pop(client.get_address(), None)
        if d:
            d.callback(None)
        else:
            self.log.warning(
                'Cannot find unanswered channel for %r' % client.get_address())

    @inlineCallbacks
    def dial_outbound(self, to_addr):
        command = self.originate_formatter.format_call(self._to_addr, to_addr)
        self.log.info("Dialing outbound via Freeswitch ESL: %r" % command)
        reply = yield self.voice_client.api(command)
        call_uuid = reply.args[1]
        self._unanswered_channels[call_uuid] = Deferred()
        returnValue(call_uuid)

    @inlineCallbacks
    def handle_outbound_message(self, message):

        client_addr = message['to_addr']
        client = self._clients.get(client_addr)

        if (self.config.supports_outbound and
            client is None and
            message.get('session_event') ==
                TransportUserMessage.SESSION_NEW):
            try:
                call_uuid = yield self.dial_outbound(client_addr)
            except FreeSwitchClientError as e:
                self.log.warning("Error connecting to client %r: %s" % (
                    client_addr, e))
                yield self.publish_nack(
                    message["message_id"],
                    "Could not make call to client %r" % (client_addr,))
            else:
                self._originated_calls[call_uuid] = message
            return

        if client is None:
            yield self.publish_nack(
                message["message_id"],
                "Client %r no longer connected" % (client_addr,))
            return

        yield self.send_outbound_message(client, message)


def get_in(d, *args, **kwargs):
    for arg in args:
        d = d.get(arg)
        if d is None:
            return kwargs.get('default')
    return d



# -*- test-case-name: vxfreeswitch.tests.test_originate -*-

"""
Utilities for creating originate FreeSwitch API calls.
"""


class OriginateMissingParameter(Exception):
    """ Raised if required originate parameters are missing. """


class OriginateFormatter(object):
    """Helper for constructing originate calls.

    This formatter uses a set of parameters to construct a template. The
    template itself is then used to construct the API call from the to and
    from addresses used by the call. The example lower down should make this
    clearer.

    :param str call_url:
        The Freeswitch sofia 'url' to originate the call to.
        Example: 'sofia/gateway/yogisip/{to_addr}'.

    :param str exten:
        The Freeswitch extension to link the call to. Example: '{from_addr}'.

    :param str cid_name:
        The caller ID name the call should appear from.  Sometimes
        required to be set by external SIP gateways such as Twilio.
        Example: 'yourapp'.

    :param str cid_num:
        The caller ID number the call should appear from.  Sometimes
        required to be set by external SIP gateways such as
        Twilio. Example: '{from_addr}'.

    :param str dialplan:
        The Freeswitch dialplan to use. Default: 'XML'.

    :param str context:
        The Freeswitch context to use. Default: 'default'.

    :param int timeout:
        The seconds to wait before timing out the dialing attempt.

    :raises OriginateParamsError:
        If required format parameters are not supplied.

    The Freeswitch originate API call format is:

      Usage: originate <call_url> <exten>|&<application_name>(<app_args>)
             [<dialplan>] [<context>] [<cid_name>] [<cid_num>] [<timeout>]

    It's rather complex and the details of each value are dependent on
    the Freeswitch configuration.

    E.g. ::

        formatter = OriginateFormatter(
            call_url='sofia/gateway/yogisip/{to_addr}',
            exten='{from_addr}',
            cid_name='vxfreeswitch',
            cid_id='{from_addr}',
        })

        api_call = formatter.format_call(
            to_addr="+1234", from_addr="+100")

        # api_call == (
        #     "originate sofia/gateway/yogisip/+1234 +100 XML default"
        #     " vxfreeswitch +100 60"
    """

    PROTO_TEMPLATE = (
        "originate %(call_url)s %(exten)s %(dialplan)s %(context)s"
        " %(cid_name)s %(cid_num)s %(timeout)s"
    )

    DEFAULT_PARAMS = {
        'dialplan': 'XML',
        'context': 'default',
        'timeout': 60,
    }

    def __init__(self, **kw):
        self.template = self.format_template(**kw)

    def format_call(self, from_addr, to_addr):
        """ Return a formatted originate call.

        :param str from_addr:
            The address the call is from.

        :param str to_addr:
            The address the call is to.

        :returns str:
            A formatted originate call for passing to Freeswitch.
        """
        return self.template.format(
            from_addr=from_addr, to_addr=to_addr)

    @classmethod
    def format_template(cls, **kw):
        """ Format a template for constructing originate calls.

        Parameters are as for the :class:`OriginateFormatter` class.
        """
        d = cls.DEFAULT_PARAMS.copy()
        d.update(kw)
        try:
            return cls.PROTO_TEMPLATE % d
        except KeyError as err:
            raise OriginateMissingParameter(
                "Missing originate parameter %s" % err)



# -*- test-case-name: vxfreeswitch.tests.test_client -*-

"""
FreeSwitch ESL API client.
"""

from twisted.internet.protocol import ClientFactory
from twisted.internet.defer import inlineCallbacks, returnValue, Deferred

from eventsocket import EventProtocol, EventError


class FreeSwitchClientProtocol(EventProtocol):
    """ Freeswitch ESL client.

    :param str auth:
        Authentication string to send to FreeSwitch.
    """

    def __init__(self, auth):
        EventProtocol.__init__(self)
        self._auth = auth
        self._connected = Deferred()

    @inlineCallbacks
    def connectionMade(self):
        if self._auth:
            yield self.auth(self._auth)
        self._connected.callback(self)


class FreeSwitchClientFactory(ClientFactory):
    """ FreeSwitch ESL client factory. """
    def __init__(self, auth=None, noisy=False):
        self.noisy = noisy
        self.auth = auth

    def protocol(self):
        return FreeSwitchClientProtocol(self.auth)


class FreeSwitchClientError(Exception):
    """ Raised when a FreeSwitch ESL command fails. """


class FreeSwitchClientReply(object):
    """ A successful reply to a FreeSwitch ESL command. """
    def __init__(self, *args):
        self.args = args

    def __repr__(self):
        return "<%s args=%r>" % (self.__class__.__name__, self.args)

    def __eq__(self, other):
        if isinstance(other, FreeSwitchClientReply):
            return self.args == other.args
        return NotImplemented


class FreeSwitchClient(object):
    """ Helper class for making simple API calls to the FreeSwitch ESL
    server.

    :type endpoint:
        Twisted client endpoint.
    :param endpoint:
        Endpoint for connecting to FreeSwitch over.

    :param str auth:
        Authentication string to send to FreeSwitch on connect.
    """
    def __init__(self, endpoint, auth=None, noisy=False):
        self.endpoint = endpoint
        self.factory = FreeSwitchClientFactory(auth=auth, noisy=noisy)

    def fallback_error_handler(self, failure):
        if failure.check(FreeSwitchClientError):
            return failure
        raise FreeSwitchClientError(str(failure.value))

    def event_error_handler(self, failure):
        if failure.check(EventError):
            err = failure.value
            ev = err.args[0]
            msg = ev.get('Reply_Text') or str(err)
            raise FreeSwitchClientError(msg)
        return failure

    def request_callback(self, ev):
        args = ev.get('Reply_Text', '').split()
        return FreeSwitchClientReply(*args)

    def api_request_callback(self, ev):
        rawresponse = ev.get('data', {}).get('rawresponse', '')
        args = rawresponse.split()
        if not (args and args[0] == "+OK"):
            msg = rawresponse or str(ev)
            raise FreeSwitchClientError(msg)
        return FreeSwitchClientReply(*args)

    @inlineCallbacks
    def _raw_with_connection(self, client, f):
        yield client._connected
        try:
            result = yield f(client)
        finally:
            yield client.transport.loseConnection()
        returnValue(result)

    def with_connection(self, f):
        """ Run a function with a connect client and then disconnect.

        :param function f:
            f(client) - the function that makes calls to the client.
        """
        d = self.endpoint.connect(self.factory)
        d.addCallback(lambda client: client._connected)
        d.addCallback(self._raw_with_connection, f)
        d.addErrback(self.event_error_handler)
        d.addErrback(self.fallback_error_handler)
        return d

    def api(self, api_call):
        def mk_call(client):
            d = client.api(api_call)
            d.addCallbacks(self.api_request_callback)
            return d
        return self.with_connection(mk_call)



from setuptools import setup, find_packages

setup(
    name="vxfreeswitch",
    version="0.1.12",
    url='http://github.com/praekelt/vumi-freeswitch-esl',
    license='BSD',
    description="A Freeswitch eventsocket transport for Vumi.",
    long_description=open('README.rst', 'r').read(),
    author='Praekelt Foundation',
    author_email='dev@praekeltfoundation.org',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'vumi',
        'confmodel>=0.2.0',
        'Twisted>=13.1.0',
        'eventsocket==0.1.4',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking',
    ],
)



""" Tests for vxfreeswitch.originate. """

from twisted.trial.unittest import TestCase

from vxfreeswitch.originate import (
    OriginateFormatter, OriginateMissingParameter)


class TestOriginateFormatter(TestCase):
    def mk_params(self, **kw):
        params = {
            'call_url': 'sofia/gateway/yogisip/{to_addr}',
            'exten': '100',
            'cid_name': 'elcid',
            'cid_num': '{from_addr}',
        }
        params.update(kw)
        params = dict((k, v) for k, v in params.items() if v is not None)
        return params

    def mk_template(self, **kw):
        return OriginateFormatter.format_template(**self.mk_params(**kw))

    def mk_formatter(self, **kw):
        return OriginateFormatter(**self.mk_params(**kw))

    def test_format_template(self):
        self.assertEqual(
            self.mk_template(),
            "originate sofia/gateway/yogisip/{to_addr}"
            " 100 XML default elcid {from_addr} 60")

    def test_format_template_missing_parameter(self):
        err = self.assertRaises(
            OriginateMissingParameter,
            self.mk_template, call_url=None)
        self.assertEqual(str(err), "Missing originate parameter 'call_url'")

    def test_init(self):
        formatter = self.mk_formatter()
        self.assertEqual(
            formatter.template,
            "originate sofia/gateway/yogisip/{to_addr}"
            " 100 XML default elcid {from_addr} 60")

    def test_init_missing_parameter(self):
        err = self.assertRaises(
            OriginateMissingParameter,
            self.mk_formatter, cid_name=None)
        self.assertEqual(str(err), "Missing originate parameter 'cid_name'")

    def test_format_call(self):
        formatter = self.mk_formatter()
        self.assertEqual(
            formatter.format_call(to_addr="+1234", from_addr="1099"),
            "originate sofia/gateway/yogisip/+1234"
            " 100 XML default elcid 1099 60")
